"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Plugin = void 0;
const fs = require("fs");
const path = require("path");
const ProgressBar = require("progress");
const typedoc_1 = require("typedoc");
const diagram_legend_1 = require("./diagram_legend");
const image_url_generator_1 = require("./image_url_generator");
const logger_1 = require("./logger");
const caching_plantuml_code_generator_1 = require("./plantuml/caching_plantuml_code_generator");
const plantuml_diagram_generator_1 = require("./plantuml/plantuml_diagram_generator");
const plugin_options_1 = require("./plugin_options");
const page_processor_1 = require("./typedoc/page_processor");
const page_section_1 = require("./typedoc/page_section");
const page_section_finder_1 = require("./typedoc/page_section_finder");
const typedoc_utils_1 = require("./typedoc/typedoc_utils");
/**
 * The UML class diagram generator plugin.
 *
 * # What does it do?
 *
 * This plugin generates an UML class diagram for classes and interfaces.
 *
 * # How does it do it?
 *
 * 1. The plugin scans through the classes and interfaces and generates PlantUML code for them.
 * 2. The plugin uses the plantuml-pipe module to generate image files from the PlantUML code.
 * 3. The plugin adds a section with the image to the page of each class and interface.
 */
class Plugin {
    constructor() {
        /** The options of this plugin. */
        this.options = new plugin_options_1.PluginOptions();
        /** Stores how many diagrams the plugin has to generate. */
        this.numberOfDiagramsToGenerate = 0;
        /** Stores the legends for the diagram of every reflection (KEY = ID of the reflection). */
        this.diagramLegends = new Map();
        /**
         * Called by the PlantUmlDiagramGenerator when a diagram images has been generated.
         * @param id ID info for the reflection for which the image was generated.
         * @param imageData The data of the generated image.
         */
        this.onImageGenerated = (id, imageData) => {
            var _a, _b, _c, _d, _e;
            let imageUrl = "";
            if (this.options.location === "local") {
                (_a = this.log) === null || _a === void 0 ? void 0 : _a.info(`Writing local image file for reflection ${id.reflection.name} ...`);
                const absoluteImageFilePath = this.writeLocalImageFileForReflection(imageData, id.reflection);
                (_b = this.log) === null || _b === void 0 ? void 0 : _b.info(`Creating local image URL for reflection ${id.reflection.name} ...`);
                imageUrl = (0, image_url_generator_1.createLocalImageFileUrl)(id.pageFilePath, absoluteImageFilePath);
            }
            else {
                (_c = this.log) === null || _c === void 0 ? void 0 : _c.info(`Creating embedded image URL for reflection ${id.reflection.name} ...`);
                imageUrl = (0, image_url_generator_1.createEmbeddedImageUrl)(imageData, this.options.format);
            }
            (_d = this.log) === null || _d === void 0 ? void 0 : _d.info(`Inserting diagram into page for reflection ${id.reflection.name} ...`);
            this.insertHierarchyDiagramIntoFile(id.pageFilePath, id.reflection, imageUrl);
            (_e = this.progressBar) === null || _e === void 0 ? void 0 : _e.tick();
        };
    }
    /**
     * Checks if the plugin is active and should generate output.
     * @returns True if the plugin is active, otherwise false.
     */
    get isActive() {
        return this.options.type === "simple" || this.options.type === "detailed";
    }
    /**
     * Checks if the plugin has anything to do.
     * @returns True if the plugin needs to generate diagrams, otherwise false.
     */
    get hasWork() {
        return this.numberOfDiagramsToGenerate > 0;
    }
    /**
     * Checks if the plugin is generating images from the generated PlantUML code.
     * @returns True if the plugin is generating images, otherwise false.
     */
    get isGeneratingImages() {
        return this.options.location === "local" || this.options.location === "embed";
    }
    /**
     * Checks if the plugin should generate diagram legends.
     * @returns True if the plugin should generate diagram legends, otherwise false.
     */
    get shouldGenerateLegends() {
        // The simple diagram type can only contain the circled chars.
        // If those are deactivated then it does not make sense to have any legends.
        if (this.options.type === "simple" && this.options.hideCircledChar) {
            return false;
        }
        else if (this.options.legendType === "only-included" || this.options.legendType === "full") {
            return true;
        }
        return false;
    }
    /**
     * Initializes the plugin.
     * @param typedoc The TypeDoc application.
     */
    initialize(typedoc) {
        this.addOptionsToApplication(typedoc);
        this.subscribeToApplicationEvents(typedoc);
    }
    /**
     * Triggered when the TypeDoc converter begins resolving a project.
     * Reads plugin parameter values.
     * @param context Describes the current state the converter is in.
     */
    onConverterResolveBegin(context) {
        this.options.readValuesFromApplication(context.converter.owner);
    }
    /**
     * Triggered when the TypeDoc converter has finished resolving a project.
     * @param context Describes the current state the converter is in.
     */
    onConverterResolveEnd(context) {
        var _a, _b, _c;
        if (this.isActive) {
            if (this.options.createVerboseOutput) {
                this.log = new logger_1.Logger();
            }
            (_a = this.log) === null || _a === void 0 ? void 0 : _a.info("Caculating number of diagrams to generate ...");
            this.numberOfDiagramsToGenerate = this.computeDiagramCount(context.project);
            (_b = this.log) === null || _b === void 0 ? void 0 : _b.info("The result is: " + this.numberOfDiagramsToGenerate.toString());
            if (this.hasWork) {
                this.plantUmlCodeGenerator = new caching_plantuml_code_generator_1.CachingPlantUmlCodeGenerator(this.options);
                if (this.isGeneratingImages) {
                    this.plantUmlDiagramGenerator = new plantuml_diagram_generator_1.PlantUmlDiagramGenerator(this.options.generatorProcessCount, this.options.format, this.onImageGenerated);
                    if (!this.options.hideProgressBar) {
                        (_c = this.log) === null || _c === void 0 ? void 0 : _c.info("Setting up progress bar ...");
                        this.progressBar = new ProgressBar(`Using ${this.options.generatorProcessCount} process(es) to add ${this.numberOfDiagramsToGenerate} class diagrams [:bar] :percent`, {
                            total: this.numberOfDiagramsToGenerate,
                            width: 40,
                        });
                    }
                }
            }
        }
    }
    /**
     * Triggered before the renderer starts rendering a project.
     * Setup image generator output directory and progress bar.
     * @param event The event emitted by the renderer class.
     */
    onRendererBegin(event) {
        if (this.isActive && this.hasWork) {
            this.outputDirectory = path.join(event.outputDirectory, "assets/");
        }
    }
    /**
     * Triggered after a document has been rendered, just before it is written to disc.
     * Generates a UML class diagram and adds it to the page of the reflection.
     * @param event The event emitted by the renderer class.
     */
    onRendererEndPage(event) {
        var _a, _b;
        if (this.isActive && this.hasWork) {
            if (this.shouldProcessReflection(event.model)) {
                if (this.shouldProcessPage(event.contents)) {
                    (_a = this.log) === null || _a === void 0 ? void 0 : _a.info(`Processing page for reflection ${event.model.name} ...`);
                    this.processPage(event);
                }
                else {
                    (_b = this.log) === null || _b === void 0 ? void 0 : _b.info(`Skipping page for reflection ${event.model.name} ...`);
                }
            }
        }
    }
    /**
     * Triggered after the renderer has written all documents.
     * Appends style data to the main CSS file.
     * @param event The event emitted by the renderer class.
     */
    onRendererEnd(event) {
        var _a, _b;
        if (this.isActive && this.hasWork) {
            (_a = this.log) === null || _a === void 0 ? void 0 : _a.info("Attaching content to CSS file ...");
            const filename = path.join(event.outputDirectory, "assets/style.css");
            let data = fs.readFileSync(filename, "utf8") +
                "\n.tsd-hierarchy-diagram .diagram { max-width: 100%; display: block; margin: 0 auto; text-align: center; }\n";
            if (this.shouldGenerateLegends) {
                data += diagram_legend_1.DiagramLegend.getCss();
            }
            fs.writeFileSync(filename, data, "utf8");
            (_b = this.log) === null || _b === void 0 ? void 0 : _b.info("DONE");
            if (this.isGeneratingImages && this.plantUmlDiagramGenerator) {
                this.plantUmlDiagramGenerator.shutdown();
            }
        }
    }
    /**
     * Adds the plugin's options to the application's options.
     * @param typedoc The TypeDoc application.
     */
    addOptionsToApplication(typedoc) {
        this.options.addToApplication(typedoc);
    }
    /**
     * Subscribes to events of the application so that the plugin can do its work
     * in the particular doc generation phases.
     * @param typedoc The TypeDoc application.
     */
    subscribeToApplicationEvents(typedoc) {
        typedoc.converter.on(typedoc_1.Converter.EVENT_RESOLVE_BEGIN, (c) => this.onConverterResolveBegin(c));
        typedoc.converter.on(typedoc_1.Converter.EVENT_RESOLVE_END, (c) => this.onConverterResolveEnd(c));
        typedoc.renderer.on(typedoc_1.RendererEvent.BEGIN, (e) => this.onRendererBegin(e));
        typedoc.renderer.on(typedoc_1.PageEvent.END, (e) => this.onRendererEndPage(e));
        typedoc.renderer.on(typedoc_1.RendererEvent.END, (e) => this.onRendererEnd(e));
    }
    /**
     * Computes how many diagrams the plugin must generate.
     * @param project The project containing the reflections for which the diagrams are generated.
     * @returns The number of diagrams to generate.
     */
    computeDiagramCount(project) {
        let count = 0;
        for (const key in project.reflections) {
            const reflection = project.reflections[key];
            if (this.shouldProcessReflection(reflection)) {
                ++count;
            }
        }
        return count;
    }
    /**
     * Checks if the plugin should process a given reflection.
     * @param reflection The reflection in question.
     * @returns True, if the plugin should process the reflection, otherwise false.
     */
    // eslint-disable-next-line class-methods-use-this
    shouldProcessReflection(reflection) {
        if (reflection instanceof typedoc_1.DeclarationReflection &&
            (reflection.kind === typedoc_1.ReflectionKind.Class || reflection.kind === typedoc_1.ReflectionKind.Interface) &&
            (0, typedoc_utils_1.reflectionIsPartOfClassHierarchy)(reflection)) {
            return true;
        }
        return false;
    }
    /**
     * Checks if the plugin should process a given page.
     * @param pageContent The content of the page.
     * @returns True, if the plugin should process the page, otherwise false.
     */
    // eslint-disable-next-line class-methods-use-this
    shouldProcessPage(pageContent) {
        if (pageContent && page_section_finder_1.PageSectionFinder.hasSection(pageContent, "Hierarchy" /* PageSections.Hierarchy */)) {
            return true;
        }
        return false;
    }
    /**
     * Processes the page by inserting a hierarchy diagram into it.
     * @param event The page event with the page data.
     */
    processPage(event) {
        var _a, _b, _c, _d, _e, _f;
        const reflection = event.model;
        (_a = this.log) === null || _a === void 0 ? void 0 : _a.info(`Generating PlantUML code for reflection ${reflection.name} ...`);
        const plantUmlLines = this.plantUmlCodeGenerator.createClassDiagramPlantUmlForReflection(reflection);
        const plantUml = plantUmlLines.join("\n");
        if (this.options.createPlantUmlFiles) {
            (_b = this.log) === null || _b === void 0 ? void 0 : _b.info(`Writing PlantUML file for reflection ${reflection.name} ...`);
            this.writePlantUmlFileForReflection(plantUml, reflection);
        }
        if (this.shouldGenerateLegends) {
            (_c = this.log) === null || _c === void 0 ? void 0 : _c.info(`Creating leged for diagram of reflection ${reflection.name} ...`);
            this.createLegendForReflection(reflection, plantUmlLines);
        }
        if (this.isGeneratingImages && this.plantUmlDiagramGenerator) {
            (_d = this.log) === null || _d === void 0 ? void 0 : _d.info(`Creating diagram image for reflection ${reflection.name} ...`);
            this.plantUmlDiagramGenerator.generate({ reflection, pageFilePath: event.filename }, plantUml + "\n");
        }
        else if (this.options.location === "remote") {
            (_e = this.log) === null || _e === void 0 ? void 0 : _e.info(`Creating remote image URL for reflection ${reflection.name} ...`);
            const imageUrl = (0, image_url_generator_1.createRemoteImageUrl)(this.options.remoteBaseUrl, plantUml, this.options.format);
            (_f = this.log) === null || _f === void 0 ? void 0 : _f.info(`Inserting diagram into page of reflection ${reflection.name} ...`);
            event.contents = this.insertHierarchyDiagramIntoContent(event.contents, reflection, imageUrl);
        }
    }
    /**
     * Creates a local file with the class diagram of a reflection.
     * @param imageData The image data of the class diagram.
     * @param reflection The reflection for which the file is written.
     * @returns The absolute path to the file that was created.
     */
    writeLocalImageFileForReflection(imageData, reflection) {
        const filename = `${reflection.name}-umlClassDiagram-${reflection.id}.${this.options.format}`;
        const absoluteFilePath = path.join(this.outputDirectory, filename);
        fs.writeFileSync(absoluteFilePath, imageData);
        return absoluteFilePath;
    }
    /**
     * Creates a file containing the PlantUML code.
     * @param plantUml The PlantUML code that should be written into the file.
     * @param reflection The reflection for which the file is written.
     */
    writePlantUmlFileForReflection(plantUml, reflection) {
        const filename = `${reflection.name}-umlClassDiagram-${reflection.id}.puml`;
        const absoluteFilePath = path.join(this.outputDirectory, filename);
        fs.writeFileSync(absoluteFilePath, plantUml, "utf8");
    }
    /**
     * Creates a legend for the reflection and stores it in the internal map.
     * @param reflection The reflection for which the legend should be generated.
     * @param plantUmlLines The PlantUML lines for the reflection's diagram.
     */
    createLegendForReflection(reflection, plantUmlLines) {
        const legend = this.options.legendType === "only-included"
            ? (0, diagram_legend_1.createDiagramLegendForPlantUml)(plantUmlLines)
            : new diagram_legend_1.DiagramLegend();
        if (this.options.hideCircledChar) {
            legend.hideTypeIcons();
        }
        if (this.options.attributeFontStyle === "italic") {
            legend.hideAbstractMemberItem();
        }
        this.diagramLegends.set(reflection.id, legend);
    }
    /**
     * Inserts the hierarchy diagram into the page file.
     * @param filePath Absolute path to the page file.
     * @param reflection The reflection for which the diagram was generated.
     * @param imageUrl The URL to the image of the diagram.
     */
    insertHierarchyDiagramIntoFile(filePath, reflection, imageUrl) {
        let fileContent = fs.readFileSync(filePath, "utf8");
        fileContent = this.insertHierarchyDiagramIntoContent(fileContent, reflection, imageUrl);
        fs.writeFileSync(filePath, fileContent, "utf8");
    }
    /**
     * Inserts the hierarchy diagram into the content string.
     * @param originalContent The content string into which the diagram is inserted.
     * @param reflection The reflection for which the diagram was generated.
     * @param imageUrl The URL to the image of the diagram.
     * @returns The updated content string with the diagram inserted into it.
     */
    insertHierarchyDiagramIntoContent(originalContent, reflection, imageUrl) {
        const page = new page_processor_1.PageProcessor(originalContent);
        const legend = this.diagramLegends.get(reflection.id);
        const hierarchyDiagramSection = (0, page_section_1.createHierarchyDiagramSection)(this.options.sectionTitle, imageUrl, reflection.name, legend && !legend.isEmpty ? legend.getHtml(this.options.visibilityStyle) : "");
        if (this.options.position === "above") {
            page.insertAboveSection("Hierarchy" /* PageSections.Hierarchy */, hierarchyDiagramSection);
        }
        else {
            page.insertBelowSection("Hierarchy" /* PageSections.Hierarchy */, hierarchyDiagramSection);
        }
        return page.content;
    }
}
exports.Plugin = Plugin;
